use std::error::Error;

use actix_web::{error, web, HttpResponse};
use futures::future;
use handlebars::Handlebars;
use lettre::{message::header::ContentType, Message, SmtpTransport, Transport};
use log::debug;
use rand::{distributions::Alphanumeric, thread_rng, Rng};
use serde::Serialize;
use serde_json::json;
use terminwahl_typen::{AppointmentSlot, Nutzer, PlannedAppointment, RequestState, Teacher};

use crate::db::{
    self,
    read::{get_slot_by_id, get_teacher_by_id},
    Pool,
};

#[derive(Serialize)]
pub struct FullAppointment {
    pub teacher: Teacher,
    pub slot: AppointmentSlot,
}

impl FullAppointment {
    pub async fn get(pool: &Pool, teacher_id: i64, slot_id: i64) -> Result<Self, Box<dyn Error>> {
        Ok(Self {
            teacher: get_teacher_by_id(pool, teacher_id)
                .await
                .expect("Teacher not found"),
            slot: get_slot_by_id(pool, slot_id).await.expect("Slot not found"),
        })
    }
}

pub async fn save_appointments_json(
    pool: web::Data<Pool>,
    mailer: web::Data<SmtpTransport>,
    handlebars: web::Data<Handlebars<'_>>,
    input: web::Json<(Vec<PlannedAppointment>, Nutzer)>,
) -> Result<HttpResponse, error::Error> {
    debug!("Extracting data");
    let (appointments, nutzer) = input.into_inner();
    debug!("Saving user");
    let nutzer_id = db::write::save_nutzer(&pool, &nutzer)
        .await
        .map_err(error::ErrorInternalServerError)?;
    debug!("Saving appointments");
    let validation_key: String = thread_rng()
        .sample_iter(&Alphanumeric)
        .take(30)
        .map(char::from)
        .collect();
    db::write::save_appointments(&pool, &appointments, nutzer_id, &validation_key)
        .await
        .map_err(error::ErrorInternalServerError)?;

    let full_appointments = future::try_join_all(appointments.into_iter().map(
        |PlannedAppointment {
             teacher_id,
             slot_id,
         }| { FullAppointment::get(&pool, teacher_id, slot_id) },
    ))
    .await
    .expect("Failed to get full list of appointments");

    let mail_result = send_confirmation_request(
        &full_appointments,
        &nutzer,
        &validation_key,
        &handlebars,
        &mailer,
    );

    Ok(HttpResponse::Ok().json(mail_result))
}

pub fn send_confirmation_request(
    appointments: &Vec<FullAppointment>,
    nutzer: &Nutzer,
    validation_key: &str,
    handlebars: &Handlebars,
    mailer: &SmtpTransport,
) -> RequestState {
    let data = json! {
    {
        "appointments": appointments,
        "nutzer": nutzer,
        "validation_key": validation_key
    }};
    debug!("{:?}", handlebars.get_templates());
    if let Ok(email_text) = handlebars.render("email_confirm", &data) {
        let email = match Message::builder()
            .from(
                "Franz Dietrich <franz.dietrich@uhlandshoehe.de>"
                    .parse()
                    .expect("Should not fail"),
            )
            .to(
                match format!("{} <{}>", nutzer.name, nutzer.email).parse() {
                    Ok(v) => v,
                    Err(_) => return RequestState::Error,
                },
            )
            .subject("Elternsprechtag: Bestätigen Sie Ihre Termine")
            .header(ContentType::TEXT_PLAIN)
            .body(
                match lettre::message::Body::new_with_encoding(
                    email_text,
                    lettre::message::header::ContentTransferEncoding::Base64,
                ) {
                    Ok(body) => body,
                    Err(_) => return RequestState::Error,
                },
            ) {
            Ok(message) => message,
            Err(_) => return RequestState::Error,
        };

        // Send the email
        match mailer.send(&email) {
            Ok(_) => RequestState::Success,
            Err(e) => {
                debug!("Failed to send: {e}");
                RequestState::Error
            }
        }
    } else {
        RequestState::Error
    }
}
